/*
Copyright(c) 1998, 1999 NOKUBI Takatsugu <knok@daionet.gr.jp>
Copyright(c) 1997 Nara Institute of Science and Technorogy.
All Rights Reserved.

$Id: ChaSen1.xs,v 1.2 1999/02/12 01:01:47 knok Exp $
*/

#ifdef __cplusplus
extern "C" {
#endif
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#ifdef __cplusplus
}
#endif

extern int Cha_optind;
int chasen_getopt_argv(char **, FILE *);

/* FILE * could be set NULL */
/* return: 0 (1: wrong option) */

int chasen_fparse(FILE *, FILE *);
int chasen_sparse(char *, FILE *);
/* return: 0 (1: end) */
char *chasen_fparse_tostr(FILE *);
char *chasen_sparse_tostr(char *);
/* return: pointer to parsed string (NULL: end) */

#define INNER_BUFSIZE   8192
#define CHA_INPUT_SIZE      8192

static unsigned char ibuf[INNER_BUFSIZE];
static unsigned char *pos;
extern char jfgets_delimiter[256];
extern int iskanji1(unsigned char *, int);
extern int jistoeuc(unsigned char *, unsigned char *);
extern void reduce_white(unsigned char *);
extern int isterminator(unsigned char *, unsigned char *);

char *sget_line(char *buffer, int bufsize, char *string)
{
  /* extern variable, ugly... */
  extern int Cha_server_mode;

  static unsigned char tmp_buf[INNER_BUFSIZE], *tmp, kanji1;
  int last;

  pos = (unsigned char *)buffer;

  tmp = tmp_buf;
  if (kanji1) {
      *tmp++ = kanji1;
      kanji1 = 0;
  }

  if(strlen(string) <= 0) {
    buffer[0] = '\0';
    pos = NULL;

    return NULL;
  }
  {
    char *lfposc;
    int lfpos;
    lfposc = strchr(string, '\n') + 1;
    if (lfposc == NULL) {
      lfpos = strlen(string);
    } else {
      lfpos = (int) (lfposc - string);
    }
    if (lfpos > bufsize) {
      memcpy(tmp, string, bufsize);
      memmove(string, (string + bufsize), strlen(string) - bufsize);
    } else {
      memcpy(tmp, string, lfpos);
      tmp[lfpos] = '\0';
      memmove(string, lfposc, strlen(lfposc) + 1);
    }
  }

  /* remove the last extra character */
  last = strlen(tmp_buf) - 1;
  if (iskanji1(tmp_buf, last)) {
      kanji1 = tmp_buf[last];
      tmp_buf[last] = 0;
  }

  /* for server mode */
  tmp = tmp_buf;
  if (Cha_server_mode)
    if (tmp_buf[0] == '.' && tmp_buf[1] == '.')
      tmp++;

  /*
   * call convertor
   * NOTE: EUC string is short than JIS string.
   *       if you want to other conversion, you must care about string length.
   */

#ifdef 0
  jistoeuc(tmp, buffer);
#else
  strcpy(buffer, tmp);
#endif

  return buffer;
}

char *jsgets(char *buffer, int bufsize, char *string)
{
  unsigned char *q;
  int count;
  int kflag;            /* kanji flag(0=not found, 1=found) */

  if(pos == NULL)
    return NULL;

  kflag = 0;
  q = (unsigned char *)buffer;

  for (count = bufsize; count > 0; count--) {
    /* line is end without '\n', long string read more */
    if(*pos == '\0')
      if(sget_line(ibuf, sizeof(ibuf), string) == NULL)
        break;

    /* KANJI */
    if(*pos >= 0x80 && *(pos+1)) {
      kflag = 1;
      count--;
      *q++ = *pos++;
      *q++ = *pos++;

      /* hit delimiter */
      if(isterminator(pos - 2, jfgets_delimiter)) {
        break;
      }
    }
    /* not KANJI */
    else {
      /* line is end */
      if(*pos == '\n') {
        if(sget_line(ibuf, sizeof(ibuf), string) == NULL)
          break;

        /* not have kanji or no space, return with this line */
        if(!kflag || count <= 0)
          break;

        /* have kanji, connect next line */
        /* double '\n' is paragraph end. so it is delimiter */
        if(*pos=='\n')
          break;

        /* If the 1st character is not KANJI, insert a space */
        if(*pos < 0xa0)
          *q++ = ' ';
      }
      else {
        *q++ = *pos++;

        /* hit delimiter */
        if(isterminator(pos - 1, jfgets_delimiter))
          break;
      }
    }

  }

  *q = '\0';

  reduce_white(buffer);

  return buffer;
}

/*PROTOTYPES: DISABLE*/

MODULE = Text::ChaSen1		PACKAGE = Text::ChaSen1

int
getopt_argv(sv,...)
	SV* sv

	PREINIT:
	char **cargs;
	int i;

	CODE:
	cargs = (char **) malloc(sizeof(char *) * items + 1);
	for (i = 0; i < items; i ++) {
		cargs[i] = SvPV(ST(i), na);
	}
	cargs[items] = NULL;
	RETVAL = chasen_getopt_argv(cargs, NULL);
	free(cargs);

	OUTPUT:
	RETVAL

SV*
sparse_tostr(sv)
	SV* sv

	PREINIT:
	char *r, *s;

	CODE:
	s = SvPV(sv, na);
	r = chasen_sparse_tostr(s);
	RETVAL = newSVpv(r, 0);

	OUTPUT:
	RETVAL

SV*
fparse_tostr(sv)
	SV* sv

	PREINIT:
	char *fname, *out;
	FILE *fp;

	CODE:
	fname = SvPV(sv, na);
	fp = fopen(fname, "r");
	if (fp == NULL) {
		RETVAL = &sv_undef;
	} else {
		RETVAL = newSVpv("", 0);
		while ((out = chasen_fparse_tostr(fp)) != NULL) {
			sv_catpv(RETVAL, out);
		}
		fclose(fp);
	}

	OUTPUT:
	RETVAL

SV*
sparse_tostr_long(sv)
	SV* sv

	PREINIT:
	char *r, *s;
	char *scopy;
	char line[CHA_INPUT_SIZE];
	long slen;

	CODE:
	s = SvPV(sv, na);
	slen = strlen(s);
	if ((scopy = malloc(slen + 1)) != NULL) {
		memcpy(scopy, s, slen + 1);
		RETVAL = newSVpv("", 0);
		pos = (unsigned char *) "";
		while (jsgets(line, sizeof(line), scopy) != NULL) {
			r = chasen_sparse_tostr(line);
			sv_catpv(RETVAL, r);
		}
		free(scopy);
	} else {
		RETVAL = &sv_undef;
	}

	OUTPUT:
	RETVAL
